from starling_sim.basemodel.algorithms.algorithm import Algorithm
from starling_sim.basemodel.agent.requests import Operation
from starling_sim.utils.utils import json_load
from starling_sim.basemodel.trace.events import StaffOperationEvent
import random


class PalZhangGCH(Algorithm):
    NAME = "PalZhangGCH"

    SCHEMA = {
        "type": "object",
        "properties": {
            "start_times": {
                "title": "Relocation start times",
                "description": "Times at which relocations will be performed",
                "type": "array",
                "items": {"type": "integer", "minimum": 0},
                "default": [25000],
            },
            "durations": {
                "title": "Relocation durations",
                "description": "Durations of each relocation operation",
                "type": "array",
                "items": {"type": "integer", "minimum": 0},
                "default": [12000],
            },
            "neighbor": {
                "title": "Neighbor evaluation method",
                "description": "Method for evaluating the station where the next relocation will take place",
                "type": "string",
                "oneOf": [
                    {"const": "util", "title": "Station with max utility (operations/time)"},
                    {"const": "nearest", "title": "Nearest station"},
                    {"const": "random", "title": "Random station"},
                ],
                "default": "util",
            },
            "threshold": {
                "type": "object",
                "properties": {
                    "min": {
                        "title": "Lower stock threshold",
                        "description": "Avoided lower stock threshold, if possible. "
                        "If integer, distance from empty station. "
                        "If between 0 and 1, the distance is computed as a fraction of "
                        "the station capacity",
                        "type": "number",
                        "default": 0.1,
                    },
                    "max": {
                        "title": "Upper stock threshold",
                        "description": "Avoided upper stock threshold, if possible. "
                        "If integer, distance from the station capacity. "
                        "If between 0 and 1, the distance is computed as a fraction of "
                        "the station capacity",
                        "type": "number",
                        "default": 0.1,
                    },
                },
            },
            "priority_threshold": {
                "title": "Priority threshold",
                "description": "Prioritised threshold when both cannot be respected",
                "type": "string",
                "oneOf": [
                    {"const": "max", "title": "Maximum"},
                    {"const": "min", "title": "Minimum"},
                ],
                "default": "max",
            },
        },
    }

    def __init__(self, simulation_model, operator, verb=False):
        super().__init__(simulation_model=simulation_model, operator=operator, verb=verb)

        # PalZhangGCH parameters

        # start times of the repositioning operations
        self.start_times = self.operator.operationParameters["start_times"]

        # duration of the operations
        self.durations = self.operator.operationParameters["durations"]

        # mode of neighbor selection
        self.neighbor = self.operator.operationParameters["neighbor"]

        self.threshold = self.operator.operationParameters["threshold"]

        self.priority_threshold = self.operator.operationParameters["priority_threshold"]

        # PalZhangGCH variables

        # demand at each station
        self.demand_dict = None

        # current simulation time
        self.current_time = None

        # time limit for the repositioning operations
        self.time_limit = None

        # current station node
        self.current_node = None

        # dict of relative number of operations needed at each station
        self.operations_needed = None

        # planning built by the algorithm
        self.planning = None

        # service vehicle (only one for this dispatch)
        self.vehicle = None

        # capacity of the repositioning vehicle
        self.vehicle_capacity = None

        # mode of the repositioning vehicle
        self.vehicle_mode = None

        # mode of the repositioning vehicle
        self.vehicle_dwell_time = None

        # depot of the repositioning vehicles
        self.depot = None

    def setup_new_run(self):
        if self.demand_dict is None:
            self.demand_dict = self.init_demand_dict()

        # consider only a single depot for now
        self.depot = list(self.operator.depotPoints.values())[0]

        self.planning = []
        self.append_depot_to_planning()

        self.current_node = self.depot.id
        self.current_time = self.sim.scheduler.now()

        if self.durations is not None:
            i = self.start_times.index(self.current_time)
            duration = self.durations[i]

            self.time_limit = self.current_time + duration

        self.operations_needed = self.compute_operations_needed()

        self.vehicle = list(self.operator.staff.values())[0]
        self.vehicle_capacity = self.vehicle.seats
        self.vehicle_mode = self.vehicle.mode
        self.vehicle_dwell_time = self.vehicle.dwellTime

    def run(self):
        # log algorithm start
        self.log_message("Start of the PalZhangGCH")

        max_operations = self.maximum_operations()

        while any(max_operations.values()):
            next_node, travel_time = self.select_next_neighbor(self.current_node, max_operations)

            if next_node in self.operator.stations:
                next_node_position = self.operator.stations[next_node].position
            else:
                next_node_position = self.operator.depotPoints[0].position
            end_time = self.current_time + travel_time + self.vehicle_dwell_time

            if self.breaking_condition(end_time):
                # TODO : what to do with current load ?
                self.log_message("End of planning due to limited operation duration")
                failed_get = 0
                failed_put = 0
                for op_needed in self.operations_needed.values():
                    if op_needed < 0:
                        failed_get -= op_needed
                    elif op_needed > 0:
                        failed_put += op_needed

                get_event = StaffOperationEvent(
                    self.sim.scheduler.now(), self.vehicle, 0, failed_get
                )
                put_event = StaffOperationEvent(
                    self.sim.scheduler.now(), self.vehicle, 0, failed_put
                )
                self.vehicle.trace_event(get_event)
                self.vehicle.trace_event(put_event)
                break

            operation = Operation(
                Operation.REPOSITIONING,
                next_node_position,
                max_operations[next_node],
                station_id=next_node,
            )

            operation.arrivalTime = int(self.current_time + travel_time)
            operation.departureTime = int(self.current_time + travel_time + self.vehicle_dwell_time)

            self.log_message("New operation {}".format(operation), lvl=10)

            # add the operation to the planning
            self.planning.append(operation)
            self.operations_needed[next_node] -= max_operations[next_node]
            self.current_time = end_time
            self.current_node = next_node
            max_operations = self.maximum_operations()

        self.append_depot_to_planning()

    def breaking_condition(self, end_time):
        # TODO : have a better breaking condition
        if self.time_limit is not None and end_time > self.time_limit:
            return True

        else:
            return False

    def compute_operations_needed(self):
        operations_needed = dict()

        for station in self.operator.stations.values():
            initial_stock = len(station.store.items)
            target_stock = self.compute_station_target_stock(station.id)
            operations_needed[station.id] = target_stock - initial_stock
            # self.log_message("So operations needed = {}".format(operations_needed[station.id]))

        return operations_needed

    def compute_station_target_stock(self, station_id):
        station = self.operator.stations[station_id]
        capacity = station.capacity

        if 0 < self.threshold["min"] < 1:
            threshold_min = int(self.threshold["min"] * capacity)
        else:
            threshold_min = self.threshold["min"]

        if 0 < self.threshold["max"] < 1:
            threshold_max = int((1 - self.threshold["max"]) * capacity)
        else:
            threshold_max = capacity - self.threshold["max"]

        # self.log_message("Station {}".format(station))
        # self.log_message("Current stock is {}".format(len(station.store.items)))
        variation = self.compute_station_variation(station_id)
        # self.log_message("|{}| < {}".format(variation, capacity - self.threshold["max"] - self.threshold["min"]))
        outrange = variation > threshold_max - threshold_min
        # self.log_message("Stock variation is {}".format(variation))
        target_stock = len(station.store.items) - variation

        if target_stock < threshold_min:
            if outrange and self.priority_threshold == "max":
                target_stock = max(0, threshold_max - variation)
            else:
                target_stock = threshold_min

        elif target_stock > threshold_max:
            if outrange and self.priority_threshold == "min":
                target_stock = min(capacity, threshold_min + variation)
            else:
                target_stock = threshold_max
        # self.log_message("So target stock is {}".format(target_stock))
        return target_stock

    def compute_station_variation(self, station_id):
        index = self.start_times.index(self.current_time)
        if index == len(self.start_times) - 1:
            horizon = self.sim.scenario["limit"]
        else:
            horizon = self.start_times[index + 1]

        station_demand = self.demand_dict[station_id]

        relevant_demand = []

        for demand in station_demand:
            if self.current_time < demand[0] < horizon:
                relevant_demand.append(demand)

        variation = 0
        for demand in relevant_demand:
            variation += demand[1]

        return variation

    def maximum_operations(self):
        max_operations = dict()

        current_load = 0
        for operation in self.planning:
            current_load -= operation.total

        for station in self.operator.stations.values():
            operation_needed = self.operations_needed[station.id]

            # positive operation corresponds to a dropoff
            if operation_needed > 0:
                maximum_operation = min(operation_needed, current_load)
            # negative operation corresponds to a pickup
            elif operation_needed < 0:
                maximum_operation = max(operation_needed, current_load - self.vehicle_capacity)
            # if no operation is needed, max operation is 0
            else:
                maximum_operation = 0

            max_operations[station.id] = maximum_operation

        return max_operations

    def select_next_neighbor(self, current_station_id, max_operations):
        # get the current station
        if current_station_id in self.operator.stations:
            current_station = self.operator.stations[current_station_id]
        else:
            current_station = self.depot

        travel_time = dict()
        # TODO : work on the stations of max_operations, to allow removing some
        for station in self.operator.stations.values():
            # ignore current station in neighbor evaluation
            if station.id == current_station_id:
                continue

            # compute travel time to other stations
            travel_time[station.id] = self.sim.environment.topologies[
                self.vehicle_mode
            ].shortest_path_length(current_station.position, station.position, None)

        # evaluate the next neighbor according to the neighbor parameter
        if self.neighbor == "nearest":
            neighbor = min(travel_time, key=travel_time.get)
        elif self.neighbor == "util":

            def utility(s):
                return float(abs(max_operations[s])) / max(1, travel_time[s])

            neighbor = max(travel_time, key=utility)
        elif self.neighbor == "random":
            neighbor = random.sample(list(travel_time.keys()), 1)[0]
        else:
            self.log_message("Unknown neighbor evaluation '{}'".format(self.neighbor))
            neighbor = None

        return neighbor, travel_time[neighbor]

    def append_depot_to_planning(self):
        operation = Operation(None, self.depot.position, 0, station_id=self.depot.id)
        self.planning.append(operation)

    def init_demand_dict(self):
        features = self.sim.dynamicInput.feature_list_from_file(
            self.sim.scenario["dynamic_input_file"]
        )
        self.sim.dynamicInput.pre_process_position_coordinates(features)
        demand_dict = {station: [] for station in self.operator.stations.keys()}

        for user_dict in features:
            origin = user_dict["properties"]["origin"]
            destination = user_dict["properties"]["destination"]
            origin_time = user_dict["properties"]["origin_time"]

            origin_station = self.sim.environment.euclidean_n_closest(
                origin, self.operator.stations.values(), 1
            )[0]

            travel_time = self.sim.environment.topologies["walk"].shortest_path_length(
                origin, origin_station.position, None
            )
            origin_station_time = origin_time + travel_time

            demand_dict[origin_station.id].append([origin_station_time, -1])

            destination_station = self.sim.environment.euclidean_n_closest(
                destination, self.operator.stations.values(), 1
            )[0]

            travel_time = self.sim.environment.topologies[
                self.operator.mode["fleet"]
            ].shortest_path_length(origin_station.position, destination_station.position, None)

            destination_station_time = origin_station_time + travel_time
            demand_dict[destination_station.id].append([destination_station_time, 1])

        for station in demand_dict.keys():
            demand_dict[station] = sorted(demand_dict[station], key=lambda x: x[0])

        return demand_dict



from starling_sim.basemodel.simulation_model import SimulationModel
from starling_sim.models.PT.output import Output
from starling_sim.basemodel.agent.vehicles.public_transport_vehicle import (
    PublicTransportVehicle,
)
from starling_sim.basemodel.agent.operators.public_transport_operator import (
    PublicTransportOperator,
)
from starling_sim.utils.constants import PUBLIC_TRANSPORT_TYPE


class Model(SimulationModel):
    """
    Model for the simulation of a Time Table Public Transport system
    """

    name = "Timetable public transport model"

    agent_type_class = {
        PUBLIC_TRANSPORT_TYPE: PublicTransportVehicle,
        "operator": PublicTransportOperator,
    }

    leaving_codes = {
        "NO_JOURNEY_FOUND": "No journey found that corresponds to request",
        "FAILED_JOURNEY": "Couldn't finish journey",
    }

    modes = {"operator": {"fleet": None}}

    output_class = Output



from starling_sim.basemodel.output.kpis import KPI

import logging
import pandas as pd


class KpiOutput:
    def __init__(self, population_names, kpi_list, kpi_name=None):
        # simulation model access
        self.sim = None

        # name of the kpi, will compose the kpi filename : <kpi_name>.csv
        if kpi_name is None:
            if isinstance(population_names, list):
                self.name = "_&_".join(population_names) + "_kpi"
            else:
                self.name = population_names + "_kpi"
        else:
            self.name = kpi_name

        # population of agent to evaluate
        self.population_names = population_names
        self.populations = None

        # list of kpi to evaluate the given agents
        self.kpi_list = kpi_list

        # output file
        self.filename = None
        self.folder = None

    def setup(self, filename, folder, simulation_model):
        """
        Setup method called during simulation setup.

        Sets the values of out file and folder, and call setup for KPIs.

        :param filename: .csv file
        :param folder:
        :param simulation_model:
        :return:
        """

        self.sim = simulation_model
        self.filename = filename
        self.folder = folder

        for kpi in self.kpi_list:
            kpi.setup(simulation_model)

        if isinstance(self.population_names, list):
            self.populations = [
                simulation_model.agentPopulation[population_name]
                for population_name in self.population_names
            ]
        else:
            self.populations = [simulation_model.agentPopulation[self.population_names]]

    def agent_kpi_dict(self, agent):
        """
        Computes the KPIs for the given agent
        by calling their update method for all its trace
        :param agent:
        :return:
        """

        indicators_dict = dict()

        # get agent trace
        events = agent.trace.eventList

        # evaluate all indicators in a single pass
        for event in events:
            for kpi in self.kpi_list:
                kpi.update(event, agent)

        # merge all completed indicators
        for kpi in self.kpi_list:
            indicators_dict.update(kpi.indicator_dict)

            # raising a warning with sphinx
            # indicators_dict = {**indicators_dict, **kpi.indicator_dict}

            # reset kpi values
            kpi.new_indicator_dict()

        # return complete indicator dict
        return indicators_dict

    def write_kpi_table(self):
        """
        Write the KPI of the population in the csv file
        obtained from out file attributes
        The KPIs evaluated are defined by the kpi_list attribute
        """

        # first row is always agent's id, then we add the kpi_list keys
        header_list = [KPI.KEY_ID]
        for kpi in self.kpi_list:
            header_list += kpi.keys

        path = self.folder + self.filename

        kpi_table = pd.DataFrame()

        # compute the kpi table for each population dict
        for population in self.populations:
            kpi_table = pd.concat([kpi_table, self.compute_population_kpi_table(population)])

        # do not generate a kpi output if the kpi table is empty
        if kpi_table.empty:
            return

        try:
            # write the dataframe into a csv file
            kpi_table.to_csv(path, sep=";", index=False, columns=header_list)

            # signal new file to output factory
            mimetype = "text/csv"
            if path.endswith("gz"):
                mimetype = "application/gzip"
            self.sim.outputFactory.new_output_file(
                path,
                mimetype,
                compressed_mimetype="text/csv",
                content="kpi",
                subject=self.name,
            )

        except KeyError as e:
            logging.warning(
                "Could not generate kpi output {}, " "error occurred : {}".format(path, e)
            )

    def compute_population_kpi_table(self, population):
        """
        Compute a kpi table for the given population dict.

        :param population: population dict {id: agent}
        :return: DataFrame containing the KPI values
        """

        df_output = pd.DataFrame()

        for agent in population.values():
            # create kpi dict for the agent
            agent_indicators = self.agent_kpi_dict(agent)

            # build a dataframe from the dict
            if isinstance(agent_indicators[KPI.KEY_ID], list):
                df = pd.DataFrame(agent_indicators)
            else:
                df = pd.DataFrame(agent_indicators, index=[0])

            # append the dataframe to the total output
            df_output = pd.concat([df_output, df])

        return df_output



from starling_sim.basemodel.topology.simple_time_weight import SimpleTimeWeight
from starling_sim.basemodel.topology.bike_weight_osm import BikeWeightOSM

import networkx as nx
from abc import ABC

# this shouldn't be here, doesn't allow adding new weights
NETWORK_WEIGHT_CLASSES = {"simple_time_weight": SimpleTimeWeight, "bike_weight_osm": BikeWeightOSM}


class Topology(ABC):
    """
    This abstract class describes a network of the simulation.

    Networks are stored as NetworkX graphs.
    """

    TIME_ATTRIBUTE = "time"
    LENGTH_ATTRIBUTE = "length"

    def __init__(self, transport_mode, weight_class=None, store_paths=False):
        """
        The constructor should not instantiate the topology data structure,
        which is done in the setup method
        """

        # transport mode described by this topology
        self.mode = transport_mode

        # networkx graph representation of the network
        self.graph = None

        # weight class
        self.weight = None
        self.init_weight(weight_class)
        self.parameters_hash = {
            self.weight.get_parameters_hash(
                self.weight.default_parameters
            ): self.weight.default_parameters
        }

        # paths storage
        self.store_paths = store_paths
        self.paths = {}
        self.shortest_path_count = 0

    # graph initialisation and setup

    def init_weight(self, weight_class):
        """
        Initialise the weight class.

        :param weight_class: weight class key
        """

        if weight_class is not None:
            if weight_class not in NETWORK_WEIGHT_CLASSES:
                raise ValueError("Unknown weight class key '{}'".format(weight_class))
        else:
            weight_class = "simple_time_weight"

        self.weight = NETWORK_WEIGHT_CLASSES[weight_class].__new__(
            NETWORK_WEIGHT_CLASSES[weight_class]
        )
        self.weight.__init__(self)

    def setup(self):
        """
        Prepare the topology for the simulation run.

        Initialise the network graph, set relevant attributes
        and compute weight values for path evaluation.
        """

        self.init_graph()

        for u, v, d in self.graph.edges(data=True):
            self.weight.pre_process_edge(u, v, d)

            self.add_time_and_length(u, v, d)

            self.compute_weights(u, v, d)

    def init_graph(self):
        """
        Initialise the network graph from any data source.
        """
        raise NotImplementedError()

    def add_time_and_length(self, u, v, d):
        """
        Set the TIME_ATTRIBUTE and LENGTH_ATTRIBUTE attributes on the graph edge.

        :param u: edge origin
        :param v: edge destination
        :param d: edge data
        """
        raise NotImplementedError()

    def compute_weights(self, u, v, d):
        """
        For each edge, set a value for each weight hash.

        :param u:
        :param v:
        :param d:
        """

        for param_hash in self.parameters_hash:
            parameters = self.parameters_hash[param_hash]
            d[param_hash] = self.weight.compute_edge_weight(u, v, d, parameters)

    # path evaluation

    def shortest_path_length(self, origin, destination, parameters):
        _, duration, _ = self.dijkstra_shortest_path_and_length(origin, destination, parameters)
        return duration

    def dijkstra_shortest_path_and_length(
        self, origin, destination, parameters, return_weight=False
    ):
        """
        Find the path from origin to destination with minimum the total weight.

        Get the hash corresponding to the parameters and call Dijkstra's algorithm
        on the corresponding weight values.

        Raises an exception if origin or destination is None,
        instead of computing all shortest path (which is what NetworkX would do)

        :param origin: origin position
        :param destination: destination position
        :param parameters: parameters defining the utility
        :param return_weight: also return the total weight

        :return: path (list of positions), duration, length
        """

        if origin is None or destination is None:
            raise ValueError("Cannot evaluate path, origin or destination is None")

        # evaluate the weight parameters
        if parameters is None:
            parameters = {}
        else:
            raise ValueError("Weight parameters are not accepted yet, please set them to None")
        for default in self.weight.default_parameters:
            if default not in parameters:
                parameters[default] = self.weight.default_parameters[default]
        param_hash = self.weight.get_parameters_hash(parameters)

        od = (origin, destination)

        if self.store_paths and od in self.paths[param_hash]:
            path, duration, length, total_weight = self.paths[param_hash][od]
        else:
            self.shortest_path_count += 1

            total_weight, path = nx.single_source_dijkstra(
                self.graph, origin, target=destination, weight=param_hash
            )

            duration, length = self.evaluate_path_duration_and_length(path)

            if self.store_paths:
                self.paths[param_hash][od] = path, duration, length, total_weight

        if return_weight:
            return path, duration, length, total_weight
        else:
            return path, duration, length

    def compute_dijkstra_path(self, origin, destination, weight):
        length, path = nx.single_source_dijkstra(
            self.graph, origin, target=destination, weight=weight
        )
        return path, length

    def evaluate_path_duration_and_length(self, path):
        total_duration = 0
        total_length = 0

        if len(path) < 1:
            raise ValueError("Cannot evaluate duration/length of path shorter than 2")

        current_position = path[0]

        for next_position in path[1:]:
            total_duration += self.get_edge_data(
                current_position, next_position, self.TIME_ATTRIBUTE
            )
            total_length += self.get_edge_data(
                current_position, next_position, self.LENGTH_ATTRIBUTE
            )
            current_position = next_position

        return total_duration, total_length

    # get graph information

    def get_positions(self):
        """
        Return the list of all topology positions

        :return: list of positions
        """

        return self.graph.nodes

    def get_edge_data(self, node1, node2, data):
        """
        Return the corresponding edge information

        :param node1: node of the topology
        :param node2: node of the topology
        :param data: requested attribute
        :return: data information of the edge
        """

        d = self.graph.get_edge_data(node1, node2)

        # TODO : check why min

        return min(attr.get(data) for attr in d.values())

    def position_localisation(self, position):
        """
        Return the localisation [lat, lon] of the position

        :param position: position in the topology

        :return: list [lat, lon]
        """
        raise NotImplementedError()

    def nearest_position(self, localisation):
        """
        Return the nearest position to given localisation (lat, lon)

        :param localisation: (lat, lon) tuple
        :return: position
        """
        raise NotImplementedError()

    def localisations_nearest_nodes(self, x_coordinates, y_coordinates, method=None):
        """
        Return the graph nodes nearest to a list of points.

        This method should allow a faster computation than multiple nearest_position calls.

        :param x_coordinates: list of X coordinates of the localisations (lon in GPS)
        :param y_coordinates: list of Y coordinates of the localisations (lat in GPS)
        :param method: name of the method used for the computation
        :return: list of nearest nodes
        """
        raise NotImplementedError()
